import logging
import os
import tempfile
from glob import glob

import pytest

from bookshelf import BookShelf
from bookshelf.notebook import NOTEBOOK_DIRECTORY, run_notebook

logger = logging.getLogger("test-notebooks")

logger.info(f"Looking for notebooks in {NOTEBOOK_DIRECTORY}")
notebooks = glob(os.path.join(NOTEBOOK_DIRECTORY, "*.py"))

notebook_names = [os.path.basename(nb)[:-3] for nb in notebooks]
logger.info(f"Found {len(notebooks)} notebooks: {notebook_names}")


@pytest.fixture()
def output_directory():
    out_dir = os.environ.get("BOOKSHELF_CACHE_LOCATION", tempfile.mkdtemp())

    yield out_dir


@pytest.mark.parametrize("notebook_path", notebooks)
def test_notebook(notebook_path, output_directory):
    # Check that:
    # * notebooks run as expected
    # * that hash matches an existing notebook

    notebook = os.path.basename(notebook_path)[:-3]
    notebook_dir = os.path.dirname(notebook_path)

    shelf = BookShelf()

    target_book = run_notebook(
        notebook,
        nb_directory=notebook_dir,
        output_directory=os.path.join(output_directory, notebook),
    )

    if shelf.is_available(name=target_book.name, version=target_book.version):
        existing_book = shelf.load(
            name=target_book.name, version=target_book.version, force=True
        )
        logger.info(f"Remote book exists. Expecting hash: {existing_book.hash()}")

        if existing_book.hash() != target_book.hash():
            raise ValueError(
                f"Hash of calculated book doesn't match the remote bookshelf "
                f"({target_book.hash()} != {existing_book.hash()})"
            )
    else:
        logger.info("Book does not exist on remote bookshelf")



import os.path
import re
import tempfile

import pytest

import bookshelf.notebook
from bookshelf.notebook import load_nb_metadata, run_notebook


def test_load_nb_metadata():
    res = load_nb_metadata("example")

    assert res.name == "example"


def test_load_nb_metadata_paths():
    exp = load_nb_metadata("example")
    assert load_nb_metadata("example")
    assert load_nb_metadata("example.yaml") == exp
    nb_dir = os.path.join(
        os.path.dirname(__file__), "..", "..", "notebooks", "example.yaml"
    )
    assert load_nb_metadata(os.path.abspath(nb_dir)) == exp


def test_run_notebook():
    with tempfile.TemporaryDirectory() as td:
        book = run_notebook("example", output_directory=str(td))

        with pytest.raises(ValueError, match=f"{td} is not empty"):
            run_notebook("example", output_directory=str(td))
        run_notebook("example", output_directory=str(td), force=True)

        assert len(book.files()) == 2
        assert sorted(os.listdir(str(td))) == sorted(
            [
                "example",  # output
                "example.ipynb",
                "example.yaml",
                "example.py",
            ]
        )


@pytest.mark.parametrize("package", ["jupytext", "papermill"])
def test_missing_deps(package):
    try:
        setattr(bookshelf.notebook, f"has_{package}", False)

        match = re.escape(
            f"{package} is not installed. Run 'pip install bookshelf[notebooks]'"
        )
        with pytest.raises(ImportError, match=match):
            run_notebook("test")
    finally:
        bookshelf.notebook.has_papermill = True
        bookshelf.notebook.has_jupytext = True



"""
Schema
"""
from typing import Any, Dict, List, Optional

from pydantic import BaseModel  # pylint: disable=no-name-in-module

Version = Optional[str]


class BookVersion(BaseModel):
    """
    Version information for a book
    """

    version: str
    url: str
    hash: str


class VolumeMeta(BaseModel):
    """
    Schema for a given Volume (A collection of Books with the same name)
    """

    name: str
    license: str  # A change in license will require a new volume
    versions: List[BookVersion]


class NotebookMetadata(BaseModel):
    """
    Schema for a given Volume (A collection of Books with the same name)
    """

    name: str
    version: str
    description: Optional[str]
    license: str
    metadata: Dict[str, Any]  # TODO: type this
    dataset: Optional[Dict[str, Any]]  # TODO: type this



import io
import json
import logging
import os
import pathlib
import re

import appdirs
import boto3
import boto3.exceptions
import moto
import pytest

from bookshelf.constants import DATA_FORMAT_VERSION
from bookshelf.errors import UnknownBook, UnknownVersion, UploadError
from bookshelf.shelf import BookShelf, LocalBook


@pytest.fixture()
def shelf(local_bookshelf):
    shelf = BookShelf(path=local_bookshelf)
    return shelf


def setup_upload_bucket(remote_bookshelf, monkeypatch):
    remote_bookshelf.mocker.get("/v0.1.0/new-package/volume.json", status_code=404)

    bucket = "test-bucket"
    prefix = "/this/prefix"
    monkeypatch.setenv("BOOKSHELF_BUCKET", bucket)
    monkeypatch.setenv("BOOKSHELF_BUCKET_PREFIX", prefix)

    conn = boto3.resource("s3", region_name="us-east-1")
    conn.create_bucket(Bucket=bucket)

    return conn


def test_local_cache(monkeypatch):
    # Override the local_bookshelf fixture
    monkeypatch.delenv("BOOKSHELF_CACHE_LOCATION")

    shelf = BookShelf()

    exp = pathlib.Path(appdirs.user_cache_dir()) / "bookshelf" / DATA_FORMAT_VERSION
    assert shelf.path == exp


def test_load_missing(shelf, remote_bookshelf):
    remote_bookshelf.mocker.get("/v0.1.0/missing/volume.json", status_code=404)

    with pytest.raises(UnknownBook, match=re.escape("No metadata for 'missing'")):
        shelf.load("missing")


def test_load_missing_version(shelf, remote_bookshelf):
    remote_bookshelf.mocker.get("/v0.1.0/test/v1.1.1/datapackage.json", status_code=404)
    with pytest.raises(UnknownVersion, match="Could not find test@v1.1.1"):
        shelf.load("test", "v1.1.1")

    with pytest.raises(UnknownVersion, match="Could not find test@v1.1.1"):
        shelf.load("test", "v1.1.1", force=True)


def test_load(remote_bookshelf, shelf):
    book = shelf.load("test")
    assert book.version == "v1.1.0"

    book = shelf.load("test", "v1.1.0")
    assert book.version == "v1.1.0"

    book = shelf.load("test", "v1.0.0")
    assert book.version == "v1.0.0"


def test_load_with_cached(remote_bookshelf, local_bookshelf):
    shelf = BookShelf(path=local_bookshelf)
    shelf.load("test")

    remote_bookshelf.mocker.reset()
    remote_bookshelf.register("test", "v1.2.1")

    shelf = BookShelf(path=local_bookshelf)
    shelf.load("test", "v1.1.0")
    assert remote_bookshelf.mocker.call_count == 0

    shelf.load("test", "v1.2.1")
    assert remote_bookshelf.mocker.call_count == 1


@moto.mock_s3
def test_publish(shelf, remote_bookshelf, monkeypatch, caplog, example_data):
    conn = setup_upload_bucket(remote_bookshelf, monkeypatch)

    book = LocalBook.create_new("new-package", "v1.1.1")
    book.add_timeseries("example", example_data)

    with caplog.at_level(logging.INFO):
        shelf.publish(book)

    # Check that files uploaded
    bucket = os.environ["BOOKSHELF_BUCKET"]
    conn.Object(bucket, "/this/prefix/new-package/v1.1.1/datapackage.json").load()
    assert (
        conn.Object(bucket, "/this/prefix/new-package/v1.1.1/datapackage.json")
        .Acl()
        .grants[1]["Permission"]
        == "READ"
    )
    conn.Object(bucket, "/this/prefix/new-package/v1.1.1/example.csv").load()

    volume_meta_contents = io.BytesIO()
    conn.Object(bucket, "/this/prefix/new-package/volume.json").download_fileobj(
        volume_meta_contents
    )
    volume_meta_contents.seek(0)
    volume_meta = json.load(volume_meta_contents)

    assert len(volume_meta["versions"]) == 1
    assert volume_meta["versions"][-1]["version"] == "v1.1.1"

    assert "Book new-package@v1.1.1 uploaded successfully" in caplog.text


@moto.mock_s3
def test_publish_new_version(
    shelf, remote_bookshelf, monkeypatch, caplog, example_data
):
    conn = setup_upload_bucket(remote_bookshelf, monkeypatch)

    book = LocalBook.create_new("test", "v1.1.1")

    with caplog.at_level(logging.INFO):
        shelf.publish(book)

    # Check that files uploaded
    bucket = os.environ["BOOKSHELF_BUCKET"]
    conn.Object(bucket, "/this/prefix/test/v1.1.1/datapackage.json").load()

    volume_meta_contents = io.BytesIO()
    conn.Object(bucket, "/this/prefix/test/volume.json").download_fileobj(
        volume_meta_contents
    )
    volume_meta_contents.seek(0)
    volume_meta = json.load(volume_meta_contents)

    assert len(volume_meta["versions"]) == 3  # 2 existing in volume.json + one extra
    assert volume_meta["versions"][-1]["version"] == "v1.1.1"

    assert "Book test@v1.1.1 uploaded successfully" in caplog.text


@moto.mock_s3
def test_publish_wrong_permissions(shelf, remote_bookshelf, monkeypatch, caplog):
    setup_upload_bucket(remote_bookshelf, monkeypatch)

    # Modify bucket used (this bucket hasn't been created)
    bucket = "test-bucket-wrong"
    monkeypatch.setenv("BOOKSHELF_BUCKET", bucket)

    caplog.set_level(logging.ERROR)

    book = LocalBook.create_new("new-package", "v1.0.0")
    with pytest.raises(UploadError, match=f"Failed to upload {book.files()[0]} to s3"):
        shelf.publish(book)

    assert "NoSuchBucket" in caplog.text
    assert bucket in caplog.text


def test_publish_existing(shelf, remote_bookshelf):
    book = LocalBook.create_new("test", "v1.0.0")

    with pytest.raises(UploadError, match="Book with the same version already exists"):
        shelf.publish(book, force=False)


@moto.mock_s3
def test_publish_existing_forced(shelf, remote_bookshelf, monkeypatch, caplog):
    setup_upload_bucket(remote_bookshelf, monkeypatch)

    book = LocalBook.create_new("new-package", "v1.0.0")

    with caplog.at_level(logging.INFO):
        shelf.publish(book, force=True)

    assert "Book new-package@v1.0.0 uploaded successfully" in caplog.text


def test_publish_extra_file(shelf, remote_bookshelf):
    remote_bookshelf.mocker.get("/v0.1.0/new-package/volume.json", status_code=404)

    book = LocalBook.create_new("new-package", "v1.0.0")
    open(book.local_fname("extra_file.txt"), "w").close()

    with pytest.raises(
        UploadError, match="Non-resource file extra_file.txt found in book"
    ):
        shelf.publish(book)


def test_is_available(shelf, remote_bookshelf):
    remote_bookshelf.mocker.get("/v0.1.0/other/volume.json", status_code=404)

    assert shelf.is_available("test", "v1.0.0")
    assert shelf.is_available("test", "v1.1.0")

    assert not shelf.is_available("test", "v1.1.1")
    assert not shelf.is_available("other", "v1.1.0")


def test_is_available_any_version(shelf, remote_bookshelf):
    remote_bookshelf.mocker.get("/v0.1.0/other/volume.json", status_code=404)

    assert shelf.is_available("test")
    assert not shelf.is_available("other")


def test_is_cached(shelf):
    book = LocalBook.create_new("test", "v1.0.0", local_bookshelf=shelf.path)

    assert shelf.is_cached(book.name, book.version)
    assert not shelf.is_cached(book.name, "v1.0.1")
    assert not shelf.is_cached("other", "v1.0.1")


def test_list_versions(shelf, remote_bookshelf):
    remote_bookshelf.mocker.get("/v0.1.0/other/volume.json", status_code=404)

    assert shelf.list_versions("test") == ["v1.0.0", "v1.1.0"]
    with pytest.raises(UnknownBook):
        shelf.list_versions("other")


@pytest.mark.xfail(reason="Not implemented")
def test_list_name(shelf):
    assert shelf.list_books() == ["test"]



"""
run CLI command
"""
import logging

import click

from bookshelf.notebook import run_notebook

logger = logging.getLogger(__name__)


@click.command("run", short_help="Run a notebook")
@click.argument("name", required=True)
@click.option(
    "-o",
    "--output",
    help="Directory to store the artifacts from running the notebook",
    required=False,
)
@click.option(
    "-f",
    "--force",
    help="Override the existing output if the output directory isn't empty",
    is_flag=True,
)
def cli(name, output, force):
    """
    Run a notebook

    This runs one of the notebooks used to generate a Book
    """
    try:
        run_notebook(name, output_directory=output, force=force)
    except Exception as exc:
        logger.error(str(exc))
        raise click.Abort() from exc



import json
import os

import pytest

from bookshelf.constants import TEST_DATA_DIR


def read_json(fname):
    fname = os.path.join(TEST_DATA_DIR, fname)
    with open(fname) as fh:
        return json.load(fh)


def read_data(fname):
    fname = os.path.join(TEST_DATA_DIR, fname)
    with open(fname) as fh:
        return fh.read()


@pytest.fixture(scope="function", autouse=True)
def remote_bookshelf(requests_mock, monkeypatch):
    monkeypatch.setenv("BOOKSHELF_REMOTE", "https://bookshelf.local/v0.1.0")

    class MockRemoteBookshelf:
        def __init__(self):
            self.mocker = requests_mock
            self.register("test", "v1.0.0")
            self.register("test", "v1.1.0")

        def register(self, name, version):
            requests_mock.get(
                f"https://bookshelf.local/v0.1.0/{name}/volume.json",
                json=read_json("v0.1.0/example/volume.json"),
            )
            requests_mock.get(
                f"https://bookshelf.local/v0.1.0/{name}/{version}/datapackage.json",
                json=read_json("v0.1.0/example/v1.0.0/datapackage.json"),
            )
            requests_mock.get(
                f"https://bookshelf.local/v0.1.0/{name}/{version}/leakage_rates_low.csv",
                text=read_data("v0.1.0/example/v1.0.0/leakage_rates_low.csv"),
            )

    bs = MockRemoteBookshelf()

    yield bs



from bookshelf import BookShelf


# Fetches from the remote bookshelf
def test_simple_book(local_bookshelf, caplog):
    with caplog.at_level(level="INFO"):
        shelf = BookShelf()
        shelf.load("rcmip-emissions", "v0.0.2")

    expected_dir = local_bookshelf / "rcmip-emissions"
    # assert (expected_dir / "volume.json").exists()
    assert (expected_dir / "v0.0.2" / "datapackage.json").exists()



"""
Functions to run/manage notebooks
"""
import logging
import os
import shutil
from typing import Optional

# pylint: disable=invalid-name
try:
    import jupytext

    has_jupytext = True
except ImportError:  # pragma: no cover
    jupytext = None
    has_jupytext = False
# pylint: disable=invalid-name
try:
    import papermill

    has_papermill = True
except ImportError:  # pragma: no cover
    papermill = None
    has_papermill = False

import yaml

from bookshelf import BookShelf, LocalBook
from bookshelf.constants import PROCESSED_DATA_DIR, ROOT_DIR
from bookshelf.schema import NotebookMetadata

logger = logging.getLogger(__name__)

NOTEBOOK_DIRECTORY = os.path.join(ROOT_DIR, "notebooks")


def load_nb_metadata(
    name: str,
    nb_directory: str = NOTEBOOK_DIRECTORY,
) -> NotebookMetadata:
    """
    Load notebook metadata

    Parameters
    ----------
    name : str
        Filename to load. Should match the notebook name (not checked)
    nb_directory: str
        If a non-absolute path is provided, it is assumed to be relative to nb_directory
    Returns
    -------
    NotebookMetadata
        Metadata about the notebook including the target package and version
    """
    if not (name.endswith(".yaml") or name.endswith(".yml")):
        name = name + ".yaml"

    if not os.path.isabs(name):
        name = os.path.join(nb_directory, name)

    with open(name) as file_handle:
        return NotebookMetadata(**yaml.safe_load(file_handle))


def run_notebook(
    name: str,
    nb_directory: str = NOTEBOOK_DIRECTORY,
    output_directory: Optional[str] = None,
    force: bool = False,
) -> LocalBook:
    """
    Run a notebook to generate a new Book

    The jupytext ``.py`` version of the notebook is used.

    The template file and configuration is copied to the output directory. The
    template ``.py`` file is then used to create a notebook which is run using
    ``papermill``. The ``local_bookshelf`` parameter is also set to the output
    directory.

    Parameters
    ----------
    name : str
        Name of the notebook
    nb_directory : str
        Directory containing the notebooks.

        This defaults to the ``notebooks/`` directory in this project
    output_directory : str
        Where the output directory will be created.

        This defaults to ``data/processing/{name}/{version}``
    force : bool
        If True, override the existing data in the output directory

    Returns
    -------
    LocalBook
        The generated book
    """
    if not has_papermill:
        raise ImportError(
            "papermill is not installed. Run 'pip install bookshelf[notebooks]'"
        )
    if not has_jupytext:
        raise ImportError(
            "jupytext is not installed. Run 'pip install bookshelf[notebooks]'"
        )

    nb_fname = os.path.join(nb_directory, f"{name}.py")
    metadata_fname = os.path.join(nb_directory, f"{name}.yaml")

    # Verify metadata
    metadata = load_nb_metadata(metadata_fname)
    if metadata.name != name:  # pragma: no cover
        raise ValueError(
            f"name in metadata does not match the name of the notebook "
            f"({metadata.name} != {name}"
        )

    if output_directory is None:
        output_directory = os.path.join(PROCESSED_DATA_DIR, name, metadata.version)
    if os.path.exists(output_directory) and os.listdir(output_directory):
        logger.warning(f"{output_directory} is not empty")
        if not force:
            raise ValueError(f"{output_directory} is not empty")
    os.makedirs(output_directory, exist_ok=True)

    # Copy required files
    logger.info(f"Copying {nb_fname} to {output_directory}")
    shutil.copyfile(nb_fname, os.path.join(output_directory, f"{name}.py"))
    logger.info(f"Copying {metadata_fname} to {output_directory}")
    shutil.copyfile(metadata_fname, os.path.join(output_directory, f"{name}.yaml"))

    # Template and run notebook
    output_nb_fname = os.path.join(output_directory, f"{name}.ipynb")
    logger.info(f"Creating notebook {output_nb_fname} from {nb_fname}")
    notebook_jupytext = jupytext.read(nb_fname)
    jupytext.write(
        notebook_jupytext,
        output_nb_fname,
        fmt="ipynb",
    )
    papermill.execute_notebook(
        output_nb_fname,
        output_nb_fname,
        parameters={"local_bookshelf": output_directory},
    )
    # Attempt to load the book from the output directory
    shelf = BookShelf(path=output_directory)
    book = shelf.load(name, metadata.version)

    logger.info(f"Notebook run successfully with hash: {book.hash()}")
    return book



"""
publish CLI command
"""
import logging
import tempfile

import click

from bookshelf.notebook import run_notebook
from bookshelf.shelf import BookShelf

logger = logging.getLogger(__name__)


@click.command("publish", short_help="Upload a book to the bookshelf")
@click.argument("name", required=True)
def cli(name):
    """
    Build and upload a Book to the Bookshelf

    Uploading a Book requires the correct AWS credentials (until an authentication
    scheme is introduced). At Climate Resource we use aws-vault for managing multiple
    sets of AWS credentials. Documentation about the different sources of authentication
    can be found here: https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-files.html

    To ensure a reproducible build, the Book is built from a notebook in an isolated
    output directory. There currently isn't any functionality to upload a pre-built Book.
    """
    with tempfile.TemporaryDirectory() as temp_dir:
        logger.info(f"Building Book in isolated environment: {temp_dir}")
        try:
            book = run_notebook(name, output_directory=temp_dir, force=False)

            shelf = BookShelf()
            shelf.publish(book)
        except Exception as exc:
            logger.error(str(exc))
            raise click.Abort() from exc



# ---
# jupyter:
#   jupytext:
#     cell_metadata_filter: -pycharm
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.14.0
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# %% [markdown]
# # CEDS
#
#
# We want to import by_region and by_region_sector

# %%
import logging
import tempfile
import zipfile

import pandas as pd
import pooch
import pycountry
import scmdata

from bookshelf import LocalBook
from bookshelf.notebook import load_nb_metadata

# %% [markdown]
# #  Initialise

# %%
logging.basicConfig(level="INFO")

# %%
metadata = load_nb_metadata("ceds")
metadata.dict()

# %% tags=["parameters"]
local_bookshelf = tempfile.mkdtemp()

# %%
local_bookshelf

# %%
book = LocalBook.create_new(
    metadata.name, version=metadata.version, local_bookshelf=local_bookshelf
)

# %% [markdown]
# #  Fetch

# %%
file_to_fetch = f"doi:{metadata.dataset['doi']}/{metadata.dataset['files'][0]}"
file_to_fetch

# %%
ceds_fname = pooch.retrieve(
    file_to_fetch,
    known_hash="95c236e7a8f7b3728453fdae8dc0e84ea8a0ec4a485dd0e96d6eff8fb40759a0",
)
ceds_fname

# %%
ceds_data = zipfile.ZipFile(ceds_fname)
[info.filename for info in ceds_data.filelist]

# %%
ceds_species = ["BC", "CH4", "CO2", "CO", "N2O", "NH3", "NMVOC", "NOx", "OC", "SO2"]
date_code = "2021_04_21"


# %% [markdown]
# # Process

# %%
def read_CEDS_format(fname: str) -> scmdata.ScmRun:
    df = pd.read_csv(ceds_data.open(fname)).rename(
        {"em": "variable", "country": "region", "units": "unit"}, axis=1
    )
    df["scenario"] = "Historical"
    df["model"] = f"CEDS ({metadata.dataset['version']})"
    df["unit"] = df["unit"] + "/yr"
    df["unit"] = df["unit"].str.replace("kt", "kt ", regex=False)
    df["variable"] = "Emissions|" + df["variable"]
    df["region"] = df["region"].str.upper()

    columns = []
    for c in df.columns:
        if c.startswith("X"):
            columns.append(int(c[1:]))
        else:
            columns.append(c)
    df.columns = columns

    return scmdata.ScmRun(df)


# %%
res = []

for species in ceds_species:
    res.append(read_CEDS_format(f"{species}_CEDS_emissions_by_country_{date_code}.csv"))
res = scmdata.run_append(res)

# %%
res["variable"] = res["variable"].str.replace("SO2", "Sulfur")
res["variable"] = res["variable"].str.replace("NMVOC", "VOC")
res["unit"] = res["unit"].str.replace("NMVOC", "VOC")

# %%
oc_emms = res.filter(variable="Emissions|OC")
oc_emms["unit"] = "kt OC/yr"
bc_emms = res.filter(variable="Emissions|BC")
bc_emms["unit"] = "kt BC/yr"
remainder_emms = res.filter(variable=["Emissions|BC", "Emissions|OC"], keep=False)

res = scmdata.run_append([oc_emms, bc_emms, remainder_emms])

# %%
res.meta[["variable", "unit"]].drop_duplicates()

# %%
res.timeseries()

# %%
# Check country codes
for c in res.get_unique_meta("region"):
    if pycountry.countries.get(alpha_3=c) is None:
        print(c)

# %%
res["region"] = res["region"].str.replace("GLOBAL", "World")

# %%
book.add_timeseries("by_country", res)


# %%
book.metadata()

# %%
